use std::cmp::Ordering::Less;
use std::fmt::{Debug, Display};
use std::ops::{Add, Sub, Mul, Div, Rem};
use std::ops::{AddAssign, SubAssign, MulAssign, DivAssign, RemAssign};
use num::Integer;

#[derive(Debug, PartialEq, Eq, PartialOrd, Clone)]
pub struct Point<T: Integer> {
    pub x: T,
    pub y: T,
}

impl<T: Integer + Default + Clone + Copy> Point<T> {
    pub fn is_zero(&self) -> bool {
        self.x.is_zero() && self.y.is_zero()
    }
    pub fn abs(&mut self) {
        if self.x.cmp(&T::zero()) == Less {
            let mut t = T::default();
            t.inc();
            t.inc();
            self.x = self.x - (self.x * t);
        }
        if self.y.cmp(&T::zero()) == Less {
            let mut t = T::default();
            t.inc();
            t.inc();
            self.y = self.y - (self.y * t);
        }
    }
    pub fn is_negative(&self) -> bool {
        if self.x.cmp(&T::zero()) == Less { return true }
        if self.y.cmp(&T::zero()) == Less { return true }
        false
    }
    pub fn distance(&self, other: &Self) -> Self {
        let mut p = Point::from((self.x-other.x, self.y-other.y));
        p.abs();
        p
    }
}

// These are defined from point_distance_value! (See macros.rs)
/*impl Point<i8> {
    pub fn distance_value(&self) -> i8 {
        if !self.is_negative() {
            return self.x + self.y;
        } else {
            let mut p = self.clone();
            p.abs();
            p.x + p.y
        }
    }
}*/

impl<T: Integer + Copy> From<T> for Point<T> {
    fn from(value: T) -> Self {
        Self { x: value, y: value }
    }
}

impl<T: Integer> From<(T, T)> for Point<T> {
    fn from(value: (T, T)) -> Self {
        Self { x: value.0, y: value.1 }
    }
}

// These are defined from point_from_string! (See macros.rs)
/*impl TryFrom<String> for Point<i8> {
    type Error = anyhow::Error;
    fn try_from(value: String) -> Result<Self, Self::Error> {
        if !value.contains("(") || !value.contains(")") || !value.contains(",") {
            anyhow::bail!("malformed point, expected \"(x, y)\", got {:?}", value);
        }
        let mut work = value.clone().trim().to_string();
        work = work.replace("(", "").replace(")", "").replace(" ", "");
        if let Some(parts) = work.split_once(",") {
            let r = parts.0.parse::<i8>();
            match r {
                Ok(x) => {
                    let r2 = parts.1.parse::<i8>();
                    match r2 {
                        Ok(y) => {
                            return Ok(Point{x, y})
                        }
                        Err(e) => {
                            return Err(anyhow::format_err!("malformed y value, {}", e));
                        }
                    }
                },
                Err(e) => {
                    return Err(anyhow::format_err!("malformed x value, {}", e));
                }
            }
        } else {
            anyhow::bail!("malformed point, expected \"(x, y)\", got {:?}", value);
        }
    }
}*/

impl<T: Integer + Default> Default for Point<T> {
    fn default() -> Self {
        Self { x: T::default(), y: T::default() }
    }
}

impl<T: Integer + Display> Display for Point<T> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str(format!("({}, {})", self.x, self.y).as_str())
    }
}

impl<T: Integer + Copy> Add<T> for Point<T> {
    type Output = Self;
    fn add(self, rhs: T) -> Self::Output {
        Self::from((self.x + rhs, self.y + rhs))
    }
}

impl<T: Integer> Add<(T, T)> for Point<T> {
    type Output = Self;
    fn add(self, rhs: (T, T)) -> Self::Output {
        Self::from((self.x+rhs.0, self.y+rhs.1))
    }
}

impl<T: Integer + Copy> Sub<T> for Point<T> {
    type Output = Self;
    fn sub(self, rhs: T) -> Self::Output {
        Self::from((self.x-rhs, self.y-rhs))
    }
}

impl<T: Integer> Sub<(T, T)> for Point<T> {
    type Output = Self;
    fn sub(self, rhs: (T, T)) -> Self::Output {
        Self::from((self.x-rhs.0, self.y-rhs.1))
    }
}

impl<T: Integer + Copy> Mul<T> for Point<T> {
    type Output = Self;
    fn mul(self, rhs: T) -> Self::Output {
        Self::from((self.x*rhs, self.y*rhs))
    }
}

impl<T: Integer> Mul<(T, T)> for Point<T> {
    type Output = Self;
    fn mul(self, rhs: (T, T)) -> Self::Output {
        Self::from((self.x*rhs.0, self.y*rhs.1))
    }
}

impl<T: Integer + Copy> Div<T> for Point<T> {
    type Output = Self;
    fn div(self, rhs: T) -> Self::Output {
        Self::from((self.x/rhs, self.y/rhs))
    }
}

impl<T: Integer> Div<(T, T)> for Point<T> {
    type Output = Self;
    fn div(self, rhs: (T, T)) -> Self::Output {
        Self::from((self.x/rhs.0, self.y/rhs.1))
    }
}

impl<T: Integer + Copy> Rem<T> for Point<T> {
    type Output = Self;
    fn rem(self, rhs: T) -> Self::Output {
        Self::from((self.x%rhs, self.y%rhs))
    }
}

impl<T: Integer> Rem<(T, T)> for Point<T> {
    type Output = Self;
    fn rem(self, rhs: (T, T)) -> Self::Output {
        Self::from((self.x%rhs.0, self.y%rhs.1))
    }
}

impl<T: Integer + AddAssign + Copy> AddAssign<T> for Point<T> {
    fn add_assign(&mut self, rhs: T) {
        self.x += rhs;
        self.y += rhs;
    }
}

impl<T: Integer + AddAssign> AddAssign<(T, T)> for Point<T> {
    fn add_assign(&mut self, rhs: (T, T)) {
        self.x += rhs.0;
        self.y += rhs.1;
    }
}

impl<T: Integer + SubAssign + Copy> SubAssign<T> for Point<T> {
    fn sub_assign(&mut self, rhs: T) {
        self.x -= rhs;
        self.y -= rhs;
    }
}

impl<T: Integer + SubAssign> SubAssign<(T, T)> for Point<T> {
    fn sub_assign(&mut self, rhs: (T, T)) {
        self.x -= rhs.0;
        self.y -= rhs.1;
    }
}

impl<T: Integer + MulAssign + Copy> MulAssign<T> for Point<T> {
    fn mul_assign(&mut self, rhs: T) {
        self.x *= rhs;
        self.y *= rhs;
    }
}

impl<T: Integer + MulAssign> MulAssign<(T, T)> for Point<T> {
    fn mul_assign(&mut self, rhs: (T, T)) {
        self.x *= rhs.0;
        self.y *= rhs.1;
    }
}

impl<T: Integer + DivAssign + Copy> DivAssign<T> for Point<T> {
    fn div_assign(&mut self, rhs: T) {
        self.x /= rhs;
        self.y /= rhs;
    }
}

impl<T: Integer + DivAssign> DivAssign<(T, T)> for Point<T> {
    fn div_assign(&mut self, rhs: (T, T)) {
        self.x /= rhs.0;
        self.y /= rhs.1;
    }
}

impl<T: Integer + RemAssign + Copy> RemAssign<T> for Point<T> {
    fn rem_assign(&mut self, rhs: T) {
        self.x %= rhs;
        self.y %= rhs;
    }
}

impl<T: Integer + RemAssign> RemAssign<(T, T)> for Point<T> {
    fn rem_assign(&mut self, rhs: (T, T)) {
        self.x %= rhs.0;
        self.y %= rhs.1;
    }
}
